#include "kinematics.h"
#include <math.h>
#include <iostream>

////platformradius
//double platformradius=10;
////thighlenght
//double thighlength =10;
////shinlenght
//double shinlength =18;
////footradius
//double footradius = 5;

#ifndef sq
#define sq(a) ((a)*(a))
#endif
#ifndef D2R
#define D2R(d) ((d)*M_PI/180.)
#endif

kinematics::kinematics()
{
    test();
}

POS kinematics::kinematic_fwd(POS p){
    double
            j0 = p.J0,
            j1 = p.J1,
            j2 = p.J2,
            y1, z1, // x1 is 0
            x2, y2, z2, x3, y3, z3,
            a1, b1, a2, b2,
            w1, w2, w3,
            denom,
            a, b, c, d;

    j0 = D2R(j0);
    j1 = D2R(j1);
    j2 = D2R(j2);

    y1 = -(p.platformradius - p.footradius + p.thighlength * cos(j0));
    z1 = -p.thighlength * sin(j0);

    y2 = (p.platformradius - p.footradius + p.thighlength * cos(j1)) * 0.5;
    x2 = y2 * sqrt(3);
    z2 = -p.thighlength * sin(j1);

    y3 = (p.platformradius - p.footradius + p.thighlength * cos(j2)) * 0.5;
    x3 = -y3 * sqrt(3);
    z3 = -p.thighlength * sin(j2);

    denom = x3 * (y2 - y1)  - x2 * (y3 - y1);

    w1 = sq(y1) + sq(z1);
    w2 = sq(x2) + sq(y2) + sq(z2);
    w3 = sq(x3) + sq(y3) + sq(z3);

    a1 = (z2-z1) * (y3-y1) - (z3-z1) * (y2-y1);
    b1 = -((w2-w1) * (y3-y1) - (w3-w1) * (y2-y1)) / 2.0;

    a2 = -(z2 - z1) * x3 + (z3 - z1) * x2;
    b2 = ((w2 - w1) * x3 - (w3 - w1) * x2) / 2.0;

    // a*z^2 + b*z + c = 0
    a = sq(a1) + sq(a2) + sq(denom);
    b = 2 * (a1 * b1 + a2 * (b2 - y1 * denom) - z1 * sq(denom));
    c = (b2 - y1 * denom) * (b2 - y1 * denom) +
            sq(b1) + sq(denom) * (sq(z1) - sq(p.shinlength));

    d = sq(b) - 4 * a * c;
    if (d < 0){
        std::cout<<"error"<<std::endl;
        p.error=1;
        return p;
    }

    p.tranZ = (-b - sqrt(d)) / (2 * a);
    p.tranX = (a1 * p.tranZ + b1) / denom;
    p.tranY = (a2 * p.tranZ + b2) / denom;

    return p;
}

POS kinematics::kinematic_inv(POS p){
    double xr, yr;

    if(inverse_J0(p, p.tranX, p.tranY, p.tranZ, &p.J0)){
        p.error=1;
        std::cout<<"error"<<std::endl;
        return p;
    }

    // now use symmetry property to get the other two just as easily...
    xr = p.tranX; yr = p.tranY;
    rotate(&xr, &yr, -2*M_PI/3);
    if(inverse_J0(p, xr, yr, p.tranZ, &p.J1)){
        p.error=1;
        std::cout<<"error"<<std::endl;
        return p;
    }

    xr = p.tranX; yr = p.tranY;
    rotate(&xr, &yr, 2*M_PI/3);
    if(inverse_J0(p, xr, yr, p.tranZ, &p.J2)){
         p.error=1;
         std::cout<<"error"<<std::endl;
         return p;
    }

    return p;
}

int kinematics::inverse_J0(POS p, double x, double y, double z, double *theta){ //Pos p is only used for robot dimensions
    double a, b, d, knee_y, knee_z;

    a = 0.5 * (sq(x) + sq(y - p.footradius) + sq(z) + sq(p.thighlength) -
               sq(p.shinlength) - sq(p.platformradius)) / z;
    b = (p.footradius - p.platformradius - y) / z;

    d = sq(p.thighlength) * (sq(b) + 1) - sq(a - b * p.platformradius);
    if (d < 0){
        std::cout<<"error"<<std::endl;
        return -1;
    }

    knee_y = (p.platformradius + a*b + sqrt(d)) / (sq(b) + 1);
    knee_z = b * knee_y - a;

    *theta = atan2(knee_z, knee_y - p.platformradius);
    *theta *= 180.0/M_PI;
    return 0;
}

void kinematics::rotate(double *x, double *y, double theta){
    double xx, yy;
    xx = *x, yy = *y;
    *x = xx * cos(theta) - yy * sin(theta);
    *y = xx * sin(theta) + yy * cos(theta);
}

void kinematics::test(){

    POS IN_fwd,OUT_fwd;

    IN_fwd.J0=0;
    IN_fwd.J0=0;
    IN_fwd.J0=0;

    OUT_fwd=kinematic_fwd(IN_fwd);

    std::cout<<"-- "<<std::endl;
    std::cout<<"Forward :"<<std::endl;
    std::cout<<"J0_:"<<OUT_fwd.J0<<std::endl;
    std::cout<<"J1:"<<OUT_fwd.J1<<std::endl;
    std::cout<<"J2:"<<OUT_fwd.J2<<std::endl;

    std::cout<<"Xtran:"<<OUT_fwd.tranX<<std::endl;
    std::cout<<"Ytran:"<<OUT_fwd.tranY<<std::endl;
    std::cout<<"Ztran:"<<OUT_fwd.tranZ<<std::endl;

    std::cout<<"Error:"<<OUT_fwd.error<<std::endl;

    std::cout<<"-- "<<std::endl;
    std::cout<<"Inverse :"<<std::endl;

    POS IN_inv,OUT_inv;

    IN_inv.tranX=2;
    IN_inv.tranY=-2;
    IN_inv.tranZ=-10;

    OUT_inv=kinematic_inv(IN_inv);

    std::cout<<"J0:"<<OUT_inv.J0<<std::endl;
    std::cout<<"J1:"<<OUT_inv.J1<<std::endl;
    std::cout<<"J2:"<<OUT_inv.J2<<std::endl;

    std::cout<<"Xtran:"<<OUT_inv.tranX<<std::endl;
    std::cout<<"Ytran:"<<OUT_inv.tranY<<std::endl;
    std::cout<<"Ztran:"<<OUT_inv.tranZ<<std::endl;

    std::cout<<"Error:"<<OUT_inv.error<<std::endl;

}






















