#ifndef ULAPI
#define ULAPI
#endif

#include "hal.h"
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <kinematics.h>
#include <iostream>


//hal realtime layer
typedef struct {
    hal_float_t *pin;
} float_data_t;

typedef struct {
    hal_bit_t *pin;
} bit_data_t;

int comp_id = hal_init("STUXNET_Cascade");
//Setup hal pins.
//Robot 0
float_data_t *Robot0_POS_CMD_J0 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval01 = hal_pin_float_new("Robot0_POS_CMD_J0",HAL_OUT,&(Robot0_POS_CMD_J0->pin),comp_id);

float_data_t *Robot0_POS_CMD_J1 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval02 = hal_pin_float_new("Robot0_POS_CMD_J1",HAL_OUT,&(Robot0_POS_CMD_J1->pin),comp_id);

float_data_t *Robot0_POS_CMD_J2 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval03 = hal_pin_float_new("Robot0_POS_CMD_J2",HAL_OUT,&(Robot0_POS_CMD_J2->pin),comp_id);

bit_data_t *Robot0_GRIPPER_IO = (bit_data_t*)hal_malloc(sizeof(bit_data_t));
int retval04 = hal_pin_bit_new("Robot0_GRIPPER_IO",HAL_OUT,&(Robot0_GRIPPER_IO->pin),comp_id);

//Robot 1
float_data_t *Robot1_POS_CMD_J0 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval05 = hal_pin_float_new("Robot1_POS_CMD_J0",HAL_OUT,&(Robot1_POS_CMD_J0->pin),comp_id);

float_data_t *Robot1_POS_CMD_J1 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval06 = hal_pin_float_new("Robot1_POS_CMD_J1",HAL_OUT,&(Robot1_POS_CMD_J1->pin),comp_id);

float_data_t *Robot1_POS_CMD_J2 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval07 = hal_pin_float_new("Robot1_POS_CMD_J2",HAL_OUT,&(Robot1_POS_CMD_J2->pin),comp_id);

bit_data_t *Robot1_GRIPPER_IO = (bit_data_t*)hal_malloc(sizeof(bit_data_t));
int retval08 = hal_pin_bit_new("Robot1_GRIPPER_IO",HAL_OUT,&(Robot1_GRIPPER_IO->pin),comp_id);

//Robot 2
float_data_t *Robot2_POS_CMD_J0 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval09 = hal_pin_float_new("Robot2_POS_CMD_J0",HAL_OUT,&(Robot2_POS_CMD_J0->pin),comp_id);

float_data_t *Robot2_POS_CMD_J1 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval10 = hal_pin_float_new("Robot2_POS_CMD_J1",HAL_OUT,&(Robot2_POS_CMD_J1->pin),comp_id);

float_data_t *Robot2_POS_CMD_J2 = (float_data_t*)hal_malloc(sizeof(float_data_t));
int retval11 = hal_pin_float_new("Robot2_POS_CMD_J2",HAL_OUT,&(Robot2_POS_CMD_J2->pin),comp_id);

bit_data_t *Robot2_GRIPPER_IO = (bit_data_t*)hal_malloc(sizeof(bit_data_t));
int retval12 = hal_pin_bit_new("Robot2_GRIPPER_IO",HAL_OUT,&(Robot2_GRIPPER_IO->pin),comp_id);

int ready = hal_ready(comp_id);
//end hal realtime layer


std::vector<POS> VPOS; //a "vector" is like a bucket that stores multiple robots in a list, including robot parameters as sub item's.

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    //kinematics(); //test function.

    /// Todo : Retrieve robots specifications from robot data file. This will be a ordinairy textfile read in.
    /// For now we setup a cascade of 50 robots.
    /// If we want to change the robot arm lenghts, etc. we read in the robot data file.
    /// Each robot in the cascade may have different dimensions.
    /// If the robot config file is ready, we loop trough the VPOS structure to set all robot arm lenghts and spec's.
    ///
    /// Todo : within the cascade floor we can calculate the active robot area, we can make logic wich robot pick's up the given xyz position to avoid interference.
    /// Todo : add minimum and maximum degrees for joints to struct.
    /// Todo : use the kdl kinematics library.
    /// Todo : add opencascade to preview the robot cascade realtime.

    //init robot cascade at program startup.. ..
    unsigned int cascade_size=50;
    VPOS.resize(cascade_size); //resize the struct vector

    //fill the struct vector with robot spec's, for now, we use the same value's for each robot in the cascade.
    //In the future this can be read in from configuration textfile. See "read_robot_config template file" for this.

    for(int i=0; i<VPOS.size(); i++){
        VPOS.at(i).J0=0;
        VPOS.at(i).J1=0;
        VPOS.at(i).J2=0;

        VPOS.at(i).tranX=0;
        VPOS.at(i).tranY=0;
        VPOS.at(i).tranZ=0;

        VPOS.at(i).footradius=5;
        VPOS.at(i).platformradius=10;
        VPOS.at(i).shinlength=18;
        VPOS.at(i).thighlength=10;

        VPOS.at(i).error=0;
    }

    std::cout<<"robot cascade size:"<<VPOS.size()<<std::endl;

    //Initialize mainwindow robot value's at program startup, it takes the current spinbox_robot_id value.
    //get a number of the current spinbox value, this for a cleaner function view below.
    unsigned int nr=ui->spinBox_robot_id->value();
    //do a forward interpolation to update mainwindow joint positions.
    VPOS.at(nr)=kinematics().kinematic_fwd(VPOS.at(nr));
    ui->lineEdit_Joint0->setText(QString::number(VPOS.at(nr).J0,'f',2));
    ui->lineEdit_Joint1->setText(QString::number(VPOS.at(nr).J1,'f',2));
    ui->lineEdit_Joint2->setText(QString::number(VPOS.at(nr).J2,'f',2));
    ui->lineEdit_error->setText(QString::number(VPOS.at(nr).error));
    //update xyz positions of mainwindow.
    ui->lineEdit_X_mdi->setText(QString::number(VPOS.at(nr).tranX,'f',2));
    ui->lineEdit_Y_mdi->setText(QString::number(VPOS.at(nr).tranY,'f',2));
    ui->lineEdit_Z_mdi->setText(QString::number(VPOS.at(nr).tranZ,'f',2));

    //start ethercat, load halfile..

    system("halcmd \-f ethercat.hal");
    system("halcmd start");

}

MainWindow::~MainWindow()
{

    hal_exit(comp_id); //when closing the program, clean up the mess.
    system("halrun -U"); //this is really important.

    delete ui;
}

void MainWindow::on_pushButton_mdi_pressed()
{
    unsigned int robot_id=ui->spinBox_robot_id->value(); //get the robot cascade number that we want to move.

    VPOS.at(robot_id).tranX=ui->lineEdit_X_mdi->text().toDouble();
    VPOS.at(robot_id).tranY=ui->lineEdit_Y_mdi->text().toDouble();
    VPOS.at(robot_id).tranZ=ui->lineEdit_Z_mdi->text().toDouble();
    VPOS.at(robot_id).error=0;

    VPOS.at(robot_id)=kinematics().kinematic_inv(VPOS.at(robot_id));

    ui->lineEdit_Joint0->setText(QString::number(VPOS.at(robot_id).J0,'f',2));
    ui->lineEdit_Joint1->setText(QString::number(VPOS.at(robot_id).J1,'f',2));
    ui->lineEdit_Joint2->setText(QString::number(VPOS.at(robot_id).J2,'f',2));
    ui->lineEdit_error->setText(QString::number(VPOS.at(robot_id).error));

    //send robot position to hal -> ethercat
    send_data(VPOS.at(robot_id),robot_id);
}

void MainWindow::on_pushButton_halshow_pressed()
{
    system("halshow &");
}

void MainWindow::send_data(POS p, unsigned nr){

    ////Todo: create a more dynamic solution.

    //realtime robot position command to "hardware abstract layer", HAL.
    if(nr==0){
        //robot ..
        *((hal_float_t *) (Robot0_POS_CMD_J0->pin)) = p.J0;
        *((hal_float_t *) (Robot0_POS_CMD_J1->pin)) = p.J1;
        *((hal_float_t *) (Robot0_POS_CMD_J1->pin)) = p.J2;

        if(ui->checkBox_gripper->checkState()==0){
            *((hal_bit_t *) (Robot0_GRIPPER_IO->pin)) = 0;
            //std::cout<<"Gripper open"<<std::endl;
        } else {
            *((hal_bit_t *) (Robot0_GRIPPER_IO->pin)) = 1;
            //std::cout<<"Gripper close"<<std::endl;
        }
    }

    if(nr==1){
        //robot ..
        *((hal_float_t *) (Robot1_POS_CMD_J0->pin)) = p.J0;
        *((hal_float_t *) (Robot1_POS_CMD_J1->pin)) = p.J1;
        *((hal_float_t *) (Robot1_POS_CMD_J1->pin)) = p.J2;

        if(ui->checkBox_gripper->checkState()==0){
            *((hal_bit_t *) (Robot1_GRIPPER_IO->pin)) = 0;
            //std::cout<<"Gripper open"<<std::endl;
        } else {
            *((hal_bit_t *) (Robot1_GRIPPER_IO->pin)) = 1;
            //std::cout<<"Gripper close"<<std::endl;
        }
    }

    if(nr==2){
        //robot ..
        *((hal_float_t *) (Robot2_POS_CMD_J0->pin)) = p.J0;
        *((hal_float_t *) (Robot2_POS_CMD_J1->pin)) = p.J1;
        *((hal_float_t *) (Robot2_POS_CMD_J1->pin)) = p.J2;

        if(ui->checkBox_gripper->checkState()==0){
            *((hal_bit_t *) (Robot2_GRIPPER_IO->pin)) = 0;
            //std::cout<<"Gripper open"<<std::endl;
        } else {
            *((hal_bit_t *) (Robot2_GRIPPER_IO->pin)) = 1;
            //std::cout<<"Gripper close"<<std::endl;
        }
    }
}








